/* GUG.pmod
 *
 * $Id: GUG.pmod,v 1.13 1998/06/11 18:23:20 nisse Exp $
 */

/*
 *    px, a Pike interface to the X Window System
 *
 *    Copyright (C) 1998, Niels Mller, Per Hedbor, Marcus Comstedt,
 *    Pontus Hagland, David Hedbor.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

/* Questions, bug fixes and bug reports can be sent to the pike
 * mailing list, pike@idonex.se, or to the athors (see AUTHORS for
 * email addresses. */

import Protocols.X;

/* Common interface for X-based widgets, returned by
 * widget->query_device_context(); */
class _xwindow
{
  object window;
  object widget;

  void show()
  {
    // werror("_xwindow->show\n");
    // werror(sprintf("Properties: %{%s,%}\n",
    //	     window->ListProperties()->name));
      
    window->Map();
  }

  void clear()
  {
    window->ClearArea(0, 0, 0, 0);
  }
  
  void set_geometry(mapping m)
  {
    window->Configure( ([ "Width" : m->width,
			      "Height" : m->height,
			      "X" : m->x, "Y": m->y ]) );
  }

  void default_resize_handler(mapping event)
  {
    widget->signal_this("resized", event->width, event->height);
  }

  void default_expose_handler(mapping event)
  {
    werror(sprintf("xwindow->expose_handler: %O\n", event));
    widget->signal_this("exposed", event->x, event->y,
			event->width, event->height,
			!event->count);
  }

  void default_button_pressed_handler(mapping event)
  {
    widget->signal("mouse_button_pressed", widget,
		   ([ "x" : event->eventX, "y" : event->eventY,
		      "time" : event->time, "button" : event->detail ]));
  }

  void default_button_released_handler(mapping event)
  {
    widget->signal("mouse_button_released", widget,
		   ([ "x" : event->eventX, "y" : event->eventY,
		      "time" : event->time, "button" : event->detail ]));
  }
  
  void create(object wd, object wt)
  {
    werror(sprintf("_xwindow->create: window = %x\n", wd->id));
    window = wd;
    widget = wt;

    /* Set default event handlers */
    window->set_event_callback("ConfigureNotify", default_resize_handler, -10);
    window->set_event_callback("Expose", default_expose_handler, -10);
    window->set_event_callback("ButtonPress",
			       default_button_pressed_handler, -10);
    window->set_event_callback("ButtonRelease",
			       default_button_released_handler, -10);
  }
}

/* Base for the Window widget */
class xwindow
{
  inherit _xwindow;

  string title;

#if 0
  void set_expose_callback(function f)
  {
    if (!expose_callback)
      {
	window->set_event_callback("Expose", expose_handler);
	window->SelectInput("Exposure");
      }
    expose_callback = f;
  }
#endif
  
#if 0
  void set_resize_callback(function f)
  {
    if (!resize_callback)
      {
	window->set_event_callback("ConfigureNotify", resize_handler);
	window->SelectInput("StructureNotify");
      }
    resize_callback = f;
  }
#endif
  
  void _set_title()
  {
    werror("xwindow->_set_title\n");
    /* Retry if TEXT can't be interned immediately */
    object ta = window->display->InternAtom("TEXT", _set_title);
    if (ta)
      {
	werror(sprintf("ta = %d, title = '%s'\n", ta->id, title));
	window->ChangeProperty(window->display->InternAtom("WM_NAME"),
			       ta, 8, title);
      }
  }
  
  void set_title(string t)
  {
    title = t;
    _set_title();
  }

  void set_min_size(int width, int height)
  {
    window->ChangeProperty(window->display->InternAtom("WM_NORMAL_HINTS"),
			   window->display->InternAtom("WM_SIZE_HINTS"),
			   32,
			   ({ 0x10, // flags = PMinSize 
				0, 0, 0, 0, // x, y, width, height
				width, height, 0, 0, // min, max
				0, 0,       // widthInc, heightInc
				0, 0, 0, 0, // aspect
				0, 0, // base width, height
				0 // winGravity
				}));
  }

  void create(object wd, object wt)
  {
    ::create(wd, wt);
    window->SelectInput("StructureNotify", "Exposure");
  }
}

class xamibutton
{
  inherit _xwindow;
  object button;
  
  object shadow_gc;
  object shine_gc;
  
  int width, height;

  function expose_callback;

  void set_geometry(int w, int h)
  {
    width = w;
    height = h;
  }

  void move(int x, int y)
  {
    werror(sprintf("xamibutton: moving to %d, %d\n", x, y));
    window->Configure( ([ "Width" : width,
			  "Height" : height,
			  "X" : x, "Y": y ]) );
  }

  void draw(int pressed)
  {
    object nw = Types.Point(0, 0);
    object ne = Types.Point(width-1, 0);
    object se = Types.Point(width-1, height-1);
    object sw = Types.Point(0, height-1);
    if (pressed)
      {
	window->DrawLine(shine_gc, 0, sw, se, ne);
	window->DrawLine(shadow_gc, 0, sw, nw, ne);
      }
    else
      {
	window->DrawLine(shadow_gc, 0, sw, se, ne);
	window->DrawLine(shine_gc, 0, sw, nw, ne);
      }
  }
  
  void redraw_handler(object ignored, int pressed, mapping event)
  {
    werror(sprintf("xbutton-> redraw_handler: %O\n", event));
    if (event)
      widget->signal_this("exposed", event->eventX, event->eventY,
			  event->width, event->height,
			  !event->count);

    if (event && event->count)
      return;

    draw(pressed);
    widget->signal(pressed ? "button_pressed" : "button_released",
		   widget);
  }

  void clicked_handler()
  {
    widget->signal("button_clicked", widget);
  }
  
  void create(object wd, object wt, mapping|object drawinfo, int n)
  {
    _xwindow::create(wd, wt);
    werror("creating gcs..\n");
    shadow_gc = window->CreateGC( (["Foreground": drawinfo->pens->shadow ]) );
    shine_gc = window->CreateGC( (["Foreground": drawinfo->pens->shine ]) );

    button = XTools.Button(window, n);
    button->redraw_callback = redraw_handler;
    button->clicked_callback = clicked_handler;

  }

#if 0
  void set_expose_callback(function f)
  {
    expose_callback = f;
  }
#endif
}
  
class xamidevice
{
  inherit Xlib.Display;
  object root;
  mapping|object drawinfo;
  object fill_gc;
  constant border_width = 1;

  mapping parse_drawinfo(array(int) data)
  {
    int i = 0;
    mapping m = ([ ]);
    while(i<sizeof(data) - 1)
      {
	int tag = data[i++];
	switch(tag)
	  {
	  case 1: // TAG_DRI_VERSION
	    m->version = data[i++];
	    break;
	  case 2: // TAG_DRI_PENS
	    {
	      int pens = data[i++];
	      if (i + pens < sizeof(data))
		return 0;
	      m->pens = ([ ]);
	      for(int j = 0; j<pens; j++)
		m->pens[_Xlib.amiwm_pens[j]] = data[i++];
	      break;
	    }
	  case 3: // TAG_DRI_FONT
	    m->font = data[i++];
	    break;
	  default:
	    werror(sprintf("xamidevice: Unknowm DrawInfo tag %d.\n",
			   tag));
	    return 0;
	  }
      }
    if ( (i != sizeof(data))
	 || !m->pens
	 || (sizeof(m->pens) < 9))
      return 0;
    return m;
  }
			   
  mapping get_drawinfo()
  {
    object name = InternAtom("AMIWM_DRAWINFO");
    object type = InternAtom("DRAWINFO");
    mapping m = root->GetProperty(name, type);

    return (m && !m->bytesAfter && parse_drawinfo(m->data))
      || ([ "pens" :
	    ([ "detail" : root->blackPixel,
	       "block" : root->whitePixel,
	       "text" : root->blackPixel,
	       "shine" : root->whitePixel,
	       "shadow" : root->blackPixel,
	       "fill" : root->whitePixel,
	       "filltext" : root->blackPixel,
	       "background" : root->whitePixel,
	       "highlighttext" : root->whitePixel,
	       "bardetail" : root->blackPixel,
	       "barblock" : root->whitePixel,
	       "bartrim" : root->blackPixel ]) ]);
  }
      
  void create()
  {
    open();
    root = DefaultRootWindow();
    drawinfo = get_drawinfo();
    fill_gc = root->CreateGC( ([ "Foreground" : drawinfo->pens->fill ]) );
  }    

  void create_window(object widget)
  {
    werror("GUG: create_window\n");
    object w =
      xwindow(root->CreateWindow(widget->x, widget->y,
				 widget->width, widget->height, 0,
				 ([ "BackPixel": root->whitePixel ])),
	      widget);
    widget->data->_context = w;
    // w->set_expose_callback(widget->refresh);
    if (widget->title)
      w->set_title(widget->title);
    w->set_min_size(widget->min_width, widget->min_height);
  }

  void show_window(object widget)
  {
    widget->data->_context->show();
  }

  void clear_window(object widget)
  {
    widget->data->_context->clear();
  }

#if 0
  void set_resize_callback(object widget, function f)
  {
    widget->data->_context->set_resize_callback(f);
  }
#endif
  
  void create_button(object widget)
  {
    werror("GUG: create_button\n");
    array a = widget->query_device_context();
    object w = a[0]->window
      ->CreateWindow(a[1], a[2],
		     widget->width, widget->height, 0,
		     ([ /* "BackPixel" : root->blackPixel */ ]) );
    object b = xamibutton(w, widget, drawinfo, widget->button_nr || 1);
    b->set_geometry(widget->width, widget->height);
    // b->set_expose_callback(widget->refresh);
    widget->data->_context = b; 
    widget->data->_ignore_parent_refresh = 1;
  }

  void show_button(object widget)
  {
    widget->data->_context->show();
  }

#if 0
  void set_button_callback(object widget, function clicked_callback)
  {
    widget->data->_context->set_button_callback(clicked_callback);
  }
#endif
  
  mapping button_geometry(mapping geometry)
  {
    mapping m = copy_value(geometry);
    m->width += 2 * border_width;
    m->height += 2 * border_width;
    return m;
  }

  mapping button_interior_geometry(mapping geometry)
  {
    mapping m = copy_value(geometry);
    m->x = border_width;
    m->y = border_width;
    m->width -= 2 * border_width;
    m->height -= 2 * border_width;
    return m;
  }

  /* FIXME: Is this test ugly? Perhaps the interface should be reorganized
   * so that set_button_geometry() is not called until after create_button()? */
  void set_button_geometry(object widget)
  {
    werror("GUG: set_button_geometry\n");
    object button = widget->data->_context;
    if (button)
      {
	werror("GUG: set_button_geometry, found device context\n");	
	array a = widget->query_device_context(1); // parent window
	button->set_geometry(widget->width, widget->height);
	button->move(a[1], a[2]);
      }
  }

  void draw_button(object widget)
  {
    widget->data->_context->draw(0);
  }
  
  void draw_solid(object widget)
  {
    array a = widget->query_device_context();
    werror(sprintf("draw_solid: device context: %O\n", a));
    a[0]->window->FillRectangles(fill_gc,
				 Types.Rectangle(a[1], a[2],
						 widget->width, widget->height));
  }

  object create_image(object widget, object image)
  {
    object xi = XImage.XImage();
    xi->set_image(image, 1);
    // xi->set_render("best");
    widget->data->_ximage = xi;
    return xi;
  }

  /* FIXME: Perhaps the interface should be reorganized so that
   * set_image_geometry() is not called until after create_image(). */
  void set_image_geometry(object widget)
  {
    if (widget->data->_ximage)
      {
	array a = widget->query_device_context();
	werror(sprintf("xamidevice->set_image_geometry: %d, %d\n",
		       a[1], a[2]));
	widget->data->_ximage->set_offset(a[1], a[2]);
      }
  }
  
  void show_image(object widget)
  {
    array a = widget->query_device_context();
    widget->data->_ximage->set_window(a[0]->window);
    widget->data->_ximage->set_offset(a[1], a[2]);
  }

  void draw_image(object widget, int x, int y, int width, int height)
  {
    widget->data->_ximage->redraw(x, y, width, height);
  }
}

object open_default_device()
{
  /* FIXME: Read .gugrc */
  object device = xamidevice();
  device->close_handler = lambda() { exit(0) ; };
  device->open(); /* blocking open */
  return device;
}
